package pds-reachability

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module type Pds_reachability_utils.Decorated_typeSource

A type for modules which carry a value type along with a few common operations on that type.

Sourcetype t
Sourceval equal : t -> t -> bool
Sourceval compare : t -> t -> int
Sourceval show : t -> string
Sourceval to_yojson : t -> Yojson.Safe.json