package pds-reachability

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module Work_type.TSource

The types module for the PDS reachability analysis.

Sourcemodule State = B.State

The decorated type of states in the PDS.

Sourcemodule Stack_element = B.Stack_element

The decorated type of stack elements in the PDS.

The decorated type of targeted dynamic pop actions in the PDS.

The decorated type of untargeted dynamic pop actions in the PDS.

Stack actions which may be performed in the PDS.

Sourceval show_stack_action : stack_action -> string
Sourceval stack_action_to_yojson : stack_action -> Yojson.Safe.json
Sourcetype node =
  1. | State_node of State.t
  2. | Intermediate_node of node * stack_action list

The decorated type of node used for reachability.

Sourcetype edge = {
  1. source : Node.t;
  2. target : Node.t;
  3. edge_action : stack_action;
}

The decorated type of edge used in reachability.